#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h> /* reiserfs uses stat */

#include <uuid/uuid.h>

#include "linux_fs.h"
#include "get_label_uuid.h"

/*
 * This code taken directly from util-linux/mount.c and util-linux/get_label_uuid.c
 * All code covered under GPL.
 * 2002-01-16 Gary Lerhaut <gary_lerhaupt@dell.com
 * - Initial implementation of partition_uuid
 */

static inline int
is_raid_partition(int fd) {
#if 0
	struct mdp_super_block mdsb;
	int n;

	/* hardcode 4096 here in various places, because that's
	   what it's defined to be.  Note that even if we used
	   the actual kernel headers, sizeof(mdp_super_t) is
	   slightly larger in the 2.2 kernel on 64-bit archs,
	   so using that wouldn't work. */
	lseek(fd, -4096, SEEK_END);	/* Ignore possible error
					   about return value overflow */
	n = 4096;
	if (sizeof(mdsb) < n)
		n = sizeof(mdsb);
	if (read(fd, &mdsb, n) != n)
		return 1;		/* error */
	return (mdsbmagic(mdsb) == MD_SB_MAGIC);
#else
	return 0;
#endif
}

static int 
valid_reiserfs_v2_super(int fd, struct reiserfs_super_block *s, int offset) {
	dev_t dev;
	struct stat st;
	int version;
	unsigned long blocksize;
	unsigned long journal_block;

	/* only 3.6.x format supers have labels or uuids */
 	if (!strncmp(s->s_magic, REISER2FS_SUPER_MAGIC_STRING,
 	            strlen(REISER2FS_SUPER_MAGIC_STRING))) {

		blocksize = assemble2le(s->s_blocksize);
		journal_block = assemble4le(s->s_journal_block);
		/* we need to make sure we aren't looking at a logged copy
		 * of a super block, so we check to make sure the offset
		 * we just read is before the offset of the first journal
		 * block.
		 */
		if ((blocksize * journal_block) < offset) {
			return 0;
		}
		return 1;
	}
 	if (!strncmp(s->s_magic, REISER2FS_JR_SUPER_MAGIC_STRING,
 	            strlen(REISER2FS_JR_SUPER_MAGIC_STRING))) {

		/* it is possible to have a relocated journal on a 
		 * 3.5.x formatted disk.  These don't support uuids or
		 * labels
		 */
		version = assemble2le(s->s_version);
		if (version != REISERFS_VERSION_2) {
			return 0;
		}

		if (fstat(fd, &st)) {
			return 0;
		}
		dev = assemble4le(s->s_journal_dev);
		if (dev == st.st_rdev) {
			/* if the journal device in the super is the 
			 * device we just read, this is a logged copy
			 * and the FS doesn't live here.
			 */ 
			return 0;
		} 
 		return 1;
	}
	return 0;
}

/* 
 * helper function to check all the possible reiserfs super locations
 */
static int 
reiserfs_super_found(int fd, struct reiserfs_super_block *reisersb) {
	int i;
	int offsets[] = {REISERFS_DISK_OFFSET_IN_BYTES,
	                   REISERFS_OLD_DISK_OFFSET_IN_BYTES, 0};

	for (i = 0 ; offsets[i] ; i++) {
		if (lseek(fd, offsets[i], SEEK_SET) == offsets[i] &&
			 read(fd, (char *)reisersb, sizeof(*reisersb)) ==
			      sizeof(*reisersb)
			&& valid_reiserfs_v2_super(fd, reisersb, offsets[i])) {
			return 1;
		}
	}
	return 0;
}


/* for now, only ext2, ext3, xfs, ocfs are supported */
int
get_label_uuid(const char *device, char **label, char *uuid) {
	int fd;
	int rv = 1;
	size_t namesize;
	struct ext2_super_block e2sb;
	struct xfs_super_block xfsb;
	struct jfs_super_block jfssb;
	struct ocfs_volume_header ovh;	/* Oracle */
	struct ocfs_volume_label olbl;
	struct reiserfs_super_block reisersb;

	fd = open(device, O_RDONLY);
	if (fd < 0)
		return 2;


	/* If there is a RAID partition, or an error, ignore this partition */
	if (is_raid_partition(fd)) {
		close(fd);
		return rv;
	}
	if (lseek(fd, 1024, SEEK_SET) == 1024
	    && read(fd, (char *) &e2sb, sizeof(e2sb)) == sizeof(e2sb)
	    && (ext2magic(e2sb) == EXT2_SUPER_MAGIC)) {
		memcpy(uuid, e2sb.s_uuid, sizeof(e2sb.s_uuid));
		namesize = sizeof(e2sb.s_volume_name);
		if ((*label = calloc(namesize + 1, 1)) != NULL)
			memcpy(*label, e2sb.s_volume_name, namesize);
		rv = 0;
	}
	else if (lseek(fd, 0, SEEK_SET) == 0
	    && read(fd, (char *) &xfsb, sizeof(xfsb)) == sizeof(xfsb)
	    && (strncmp(xfsb.s_magic, XFS_SUPER_MAGIC, 4) == 0)) {
		memcpy(uuid, xfsb.s_uuid, sizeof(xfsb.s_uuid));
		namesize = sizeof(xfsb.s_fname);
		if ((*label = calloc(namesize + 1, 1)) != NULL)
			memcpy(*label, xfsb.s_fname, namesize);
		rv = 0;
	}
	else if (lseek(fd, 0, SEEK_SET) == 0
	    && read(fd, (char *) &ovh, sizeof(ovh)) == sizeof(ovh)
	    && (strncmp(ovh.signature, OCFS_MAGIC, sizeof(OCFS_MAGIC)) == 0)
	    && (lseek(fd, 512, SEEK_SET) == 512)
	    && read(fd, (char *) &olbl, sizeof(olbl)) == sizeof(olbl)) {
		uuid[0] = '\0';
		namesize = ocfslabellen(olbl);
		if ((*label = calloc(namesize + 1, 1)) != NULL)
			memcpy(*label, olbl.label, namesize);
		rv = 0;
	}
	else if (lseek(fd, JFS_SUPER1_OFF, SEEK_SET) == JFS_SUPER1_OFF
	    && read(fd, (char *) &jfssb, sizeof(jfssb)) == sizeof(jfssb)
	    && (strncmp(jfssb.s_magic, JFS_MAGIC, 4) == 0)) {
		if (assemble4le(jfssb.s_version) == 1
		  && strncmp(jfssb.s_fpack, jfssb.s_label, 11)) {
			/* old (OS/2 compatible) jfs filesystems don't
			   have UUIDs and only have a very small label. */
		        /* strncmp hack contributed by Dave Kleikamp, JFS
			   maintainer, to really distringuish between 
			   version 1 and version 2. */
			memset(uuid, 0, 16);
			namesize = sizeof(jfssb.s_fpack);
			if ((*label = calloc(namesize + 1, 1)) != NULL)
				memcpy(*label, jfssb.s_fpack, namesize);
			rv = 3;
		} else {
			memcpy(uuid, jfssb.s_uuid, sizeof(jfssb.s_uuid));
			namesize = sizeof(jfssb.s_label);
			if ((*label = calloc(namesize + 1, 1)) != NULL)
			    memcpy(*label, jfssb.s_label, namesize);
			rv = 0;
		}
	}
	else if (reiserfs_super_found(fd, &reisersb)) {
	        memcpy(uuid, reisersb.s_uuid, sizeof(reisersb.s_uuid));
	        namesize = sizeof(reisersb.s_label);
	        if ((*label = calloc(namesize + 1, 1)) != NULL)
	                 memcpy(*label, reisersb.s_label, namesize);
	        rv = 0;
	}

	close(fd);
	return rv;
}

int 
main(int argc, char *argv[])
{
	int rc, k;
	char *file_name = 0;
	char *device = NULL;
	char *device_label = NULL;
	char device_uuid_text[40];
	unsigned char device_uuid[16];

	for (k = 1; k < argc; ++k) {
		if (*argv[k] != '-')
			file_name = argv[k];
	}
	if (file_name == 0) {
		fprintf(stderr,"Usage: %s <partition>\n", argv[0]);
      		exit(1);
	}

	device = file_name;

	rc = get_label_uuid(device, &device_label, device_uuid);
	if (device_label)
	        free(device_label);
	if (rc == 0) {
	        memset(device_uuid_text,0,sizeof(device_uuid_text));
	        uuid_unparse(device_uuid, device_uuid_text);
	        fprintf(stdout,"%s\n",device_uuid_text);
	} else if (rc == 2) {
  	        fprintf(stderr,"partition_uuid: invalid permissions or device does not exist\n");
	} else {
	        fprintf(stderr,"partition_uuid: partition does not appear to have a UUID\n");
	}

	exit (rc);
}
