#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <scsi/scsi.h>
#include <linux/major.h>
#include <linux/cciss_ioctl.h>

static char *version="scsi_unique_id.c v. 0.99";


/*  This program originally derived from and inspired by Doug 
 *  Gilbert's scsi_inquiry.c, part of sg_utils package.  
 *  See http://www.torque.net/sg
 *  Copyright (C) 2001 Compaq Computer Corporation 
 *  Copyright (C) 1999 D. Gilbert
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 */


/*  This program derived from Steve Cameron's <smcameron@yahoo.com> 
 *  implementation of scsi_serialno as part of his scsimap package.
 *  Original derivation from Doug Gilbert as stated above.
 *  Modifications by Gary Lerhaupt <gary_lerhaupt@dell.com>
 *  Copyright (C) 2002-2004 Dell Computer Corporation
 */

/* Added vpd check (do_scsi_vpdp_inquiry) coded by Dan Clark <dclark@aracnet.com>
 * Jan 06, 2004 - version 0.99
 */

typedef struct my_scsi_ioctl_command {
	unsigned int inlen;  /* _excluding_ scsi command length */
	unsigned int outlen;
	unsigned char data[1];  /* was 0 but that's not ISO C!! */
		/* on input, scsi command starts here then opt. data */
} My_Scsi_Ioctl_Command;

#define OFF (2 * sizeof(unsigned int))

#ifndef SCSI_IOCTL_SEND_COMMAND
#define SCSI_IOCTL_SEND_COMMAND 1
#endif

#define INQUIRY_CMD     0x12
#define INQUIRY_CMDLEN  6
#define INQUIRY_REPLY_LEN 96

#define UNKNOWN_TYPE -1
#define SCSI_TYPE 0

int
stat_device(int fd, unsigned int *rdev, int *devtype)
{
	struct stat statbuf;
	int res, to;

	*devtype = UNKNOWN_TYPE;
	res = fstat(fd, &statbuf);
	if (res != 0) {
		fprintf(stderr, "Cannot stat, %s\n", strerror(errno));
		return -1;
	}

	*rdev = statbuf.st_rdev;

	if (! S_ISBLK(statbuf.st_mode) ) {
		fprintf(stderr, "Not a block device!\n");
		return -1;
	}
	// printf("st_rdev = 0x%08x\n", statbuf.st_rdev);

    // Don't worry, being very careful not to write to a none-scsi file ... 
    if (*devtype == UNKNOWN_TYPE) {
    	res = ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &to);
    	if (res < 0) {
        	// perror("ioctl on scsi device, error"); 
        	return -1;
    	}
	else
	*devtype = SCSI_TYPE;
    }

    return 0;
}


int
generic_scsi_inquiry(int fd, unsigned char evpd, unsigned char page,
	unsigned char *buf, unsigned char  buflen) 
{
	unsigned char inqCmdBlk[INQUIRY_CMDLEN] = 
		{INQUIRY_CMD, evpd, page, 0, INQUIRY_REPLY_LEN, 0};
	int rc, len;
	unsigned char *inqbuf;
	unsigned char *buffp;
	My_Scsi_Ioctl_Command * ishp;

	inqbuf = malloc(OFF + sizeof(inqCmdBlk) + 512);
	memset(inqbuf, 0, OFF + sizeof(inqCmdBlk) + 512); 
	buffp = inqbuf + OFF;

	ishp = (My_Scsi_Ioctl_Command *) inqbuf;
	ishp->inlen = 0;
	ishp->outlen = INQUIRY_REPLY_LEN;
	memcpy(buffp, inqCmdBlk, INQUIRY_CMDLEN);
	rc = ioctl(fd, SCSI_IOCTL_SEND_COMMAND, inqbuf);
	if (rc != 0) {
		free(inqbuf);
		return rc;
	}
	len = INQUIRY_REPLY_LEN < buflen ?  INQUIRY_REPLY_LEN : buflen;
	memcpy(buf, buffp, len); 
	free(inqbuf);
	return len;
}

int
check_device_online(int fd, int rdev)
{
	int len, peripheral_qualifier;
	unsigned char buf[256];

	memset(buf, 0, 256);
	len = generic_scsi_inquiry(fd, 1, 0x80, buf, 255);
	peripheral_qualifier = buf[0] & 224;

	if (peripheral_qualifier == 32) {
		return -1;
	} else {
		return 0;
	}
}
	
int
do_scsi_info_inquiry(int fd, int rdev)
{
	int len, i;
	unsigned char buf[256];

	memset(buf, 0, 256);
	len = generic_scsi_inquiry(fd, 0, 0, buf, 255);
	if (len < 0) return len;
	printf("model: ");
	for (i=8;i<32;i++)
	        if (isprint(buf[i])) {
		     printf("%c", buf[i]);
	        }
	printf("\n");
	
	return 0;
}

	
int
do_scsi_83_inquiry(int fd, int rdev)
{
	int len, i, j, id_length, id_type, id_association;
	unsigned char buf[256];

	memset(buf, 0, 256);
	len = generic_scsi_inquiry(fd, 1, 0x83, buf, 255);
	if (len < 0) return len;

	if (buf[1] != 0x83) {
	        fprintf(stderr,"page83: No page83 information found.\n");
		return -1;
	} else {
		// Print out the Page 83 Header
		//printf("\n%02x %02x %02x %02x [Page Header] [Page Length=%d]\n", buf[0], buf[1], buf[2], buf [3], buf[3]);

		for (i=4; i < buf[3] + 4; i++) {
			// Get info about the ID
			id_type = buf[i+1] & 0x0f;
			id_association = buf[i+1] & 0x30;
			id_length = buf[i+3];

			// Print out the Indentifier Header
			//printf("%02x %02x %02x %02x [ID Header]", buf[i], buf[i+1], buf[i+2], buf[i+3]);
			//printf(" [Type=%d]\n", id_type);
			

			// Print out the Identifier
			i = i + 4;
			if (id_association == 0x00) {
			       printf("page83 type%d: ", id_type);
			       for (j=0; j < id_length; j++) {
			              printf("%02x", buf[i+j]);
			       }
			       printf("\n");
			} else {
			       j = id_length;
			       fprintf(stderr,"page83: One type%d identifier is not constant (non-zero association value).\n", id_type);
			}
			
			  // Increment i but subtract 1 to compensate for j loop
			i=i+j-1;
		}
	
		return 0;
	}
}       

int
do_scsi_80_inquiry(int fd, int rdev)
{
	int len, i, id_length;
	unsigned char buf[256];

	memset(buf, 0, 256);
	len = generic_scsi_inquiry(fd, 1, 0x80, buf, 255);
	if (len < 0) return len;

	if (buf[1] != 0x80) {
		fprintf(stderr,"page80: No page80 information found.\n");
		return -1;
	} else {
		// Print out the Page 80 Header
		//printf("%02x %02x %02x %02x [Page Header] [Page Length=%d]\n", buf[0], buf[1], buf[2], buf [3], buf[3]);

		// Print out the Serial #
		id_length = buf[3];
		printf("page80: ");
		for (i=4; i < id_length + 4; i++) {
			printf("%02x", buf[i]);
		}
		printf("\n");	
		return 0;
	}
}

/*
 * inquire for the vital product data pages 
 * determine if 0x80 and 0x83 inquiries are supported or not
 * then make inquiry of pages that are supported by device/driver
 * 0x00 supported vital product data pages
 * 0x80 unit serial number page
 * 0x83 device identification page
 */
int
do_scsi_vpdp_inquiry(int fd, int rdev)
{
        int len, i;
        unsigned char buf[256];
        ushort vpdp_cnt;        /* count of vpdp pages */

        memset(buf, 0, 256);
        len = generic_scsi_inquiry(fd, 1, 0, buf, 255);
        if (len < 0) return len;
        vpdp_cnt = (ushort)buf[3];
        for(i = 0; i < vpdp_cnt; i++) {
                ushort vpdp_type;
                vpdp_type = buf[4 + i];
                /* printf("0x%02x ", vpdp_type); */
                switch(vpdp_type) {
                case 0x80:
                        // Get SCSI page80 unite serial number page
                        do_scsi_80_inquiry(fd, rdev);
                        break;
                case 0x83:
                        // Get SCSI page83 device identification page
                        do_scsi_83_inquiry(fd, rdev);
                        break;
                default:
                        break;
                }
        }
        /* printf("\n"); */
        return 0;
}

int 
main(int argc, char *argv[])
{
	int fd, rc, k, devtype;
	unsigned int rdev;
	char * file_name = 0;

	if (argc >= 2 && strcmp(argv[1] , "-v")==0) {
		fprintf(stderr, "%s\n", version);
		exit (0);
	}
	for (k = 1; k < argc; ++k) {
		if (*argv[k] != '-')
			file_name = argv[k];
	}
	if (0 == file_name) {
		fprintf(stderr,"Usage: '%s <device>'\n", argv[0]);
		fprintf(stderr,"    For example: %s /dev/sda\n", argv[0]);
		exit(1);
	}

	fd = open(file_name, O_RDWR);
	if (fd < 0) {
		if ((EROFS == errno) || (EACCES == errno)) {
			fd = open(file_name, O_RDONLY);
			if (fd < 0) {
				fprintf(stderr, "%s: cannot open %s: %s\n",
					argv[0], file_name, strerror(errno));
				exit(1);
			}
		} else {
			fprintf(stderr, "%s: cannot open %s: %s\n",
				argv[0], file_name, strerror(errno));
			exit(1);
		}
	}

	if (stat_device(fd, &rdev, &devtype) != 0) {
                fprintf(stderr, "%s: Unknown devtype on %s\n", argv[0], file_name);
		exit(1);
	}	

	//Check if device is offline
	if ( check_device_online(fd, rdev) != 0) {
		fprintf(stderr, "%s: Device %s is marked offline\n", argv[0], file_name);
		exit(1);	
	}
	
	// Print out generic information about this device
	do_scsi_info_inquiry(fd, rdev);	

	// inquire for the optional vital product data pages 
	do_scsi_vpdp_inquiry(fd, rdev);

	// Get SCSI page83 info
	//do_scsi_83_inquiry(fd, rdev);
	//
	// Get SCSI page80 info
	//do_scsi_80_inquiry(fd, rdev);
	
	rc = close(fd);
	if (rc < 0) {
		perror("scsi_inquiry: close error");
		return 1;
	}
	
	return 0;
}
